@@ -1,27 +1,7 @@
-/* PipeWire
- *
- * Copyright © 2021 Sanchayan Maity <sanchayan@asymptotic.io>
- * Copyright © 2022 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2021 Sanchayan Maity <sanchayan@asymptotic.io> */
+/* SPDX-FileCopyrightText: Copyright © 2022 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <string.h>
 #include <stdio.h>
@@ -63,16 +43,16 @@
  * - `stream.props`: Extra properties for the local stream.
  *
  * When `tunnel.mode` is `capture`, a capture stream on the default source is
- * created. Samples read from the pipe will be the contents of the captured source.
+ * created. The samples captured from the source will be written to the pipe.
  *
- * When `tunnel.mode` is `sink`, a sink node is created. Samples read from the
- * pipe will be the samples played on the sink.
+ * When `tunnel.mode` is `sink`, a sink node is created. Samples played on the
+ * sink will be written to the pipe.
  *
  * When `tunnel.mode` is `playback`, a playback stream on the default sink is
- * created. Samples written to the pipe will be played on the sink.
+ * created. The samples read from the pipe will be played on the sink.
  *
- * When `tunnel.mode` is `source`, a source node is created. Samples written to
- * the pipe will be made available to streams connected to the source.
+ * When `tunnel.mode` is `source`, a source node is created. Samples read from
+ * the the pipe will be made available on the source.
  *
  * When `pipe.filename` is not given, a default fifo in `/tmp/fifo_input` or
  * `/tmp/fifo_output` will be created that can be written and read respectively,
@@ -123,7 +103,7 @@
 
 #define NAME "pipe-tunnel"
 
-#define DEFAULT_CAPTURE_FILENAME 	"/tmp/fifo_input"
+#define DEFAULT_CAPTURE_FILENAME	"/tmp/fifo_input"
 #define DEFAULT_PLAYBACK_FILENAME	"/tmp/fifo_output"
 
 #define DEFAULT_FORMAT "S16"
@@ -131,21 +111,24 @@
 #define DEFAULT_CHANNELS 2
 #define DEFAULT_POSITION "[ FL FR ]"
 
+#define RINGBUFFER_SIZE		(1u << 22)
+#define RINGBUFFER_MASK		(RINGBUFFER_SIZE-1)
+
 PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 #define PW_LOG_TOPIC_DEFAULT mod_topic
 
-#define MODULE_USAGE	"[ remote.name=<remote> ] "				\
-			"[ node.latency=<latency as fraction> ] "		\
-			"[ node.name=<name of the nodes> ] "			\
-			"[ node.description=<description of the nodes> ] "	\
-			"[ target.object=<remote node target name or serial> ] "\
-			"[ audio.format=<sample format> ] "			\
-			"[ audio.rate=<sample rate> ] "				\
-			"[ audio.channels=<number of channels> ] "		\
-			"[ audio.position=<channel map> ] "			\
-			"[ tunnel.mode=capture|playback|sink|source "		\
-			"[ pipe.filename=<filename> ]"				\
-			"[ stream.props=<properties> ] "
+#define MODULE_USAGE	"( remote.name=<remote> ) "				\
+			"( node.latency=<latency as fraction> ) "		\
+			"( node.name=<name of the nodes> ) "			\
+			"( node.description=<description of the nodes> ) "	\
+			"( target.object=<remote node target name or serial> ) "\
+			"( audio.format=<sample format> ) "			\
+			"( audio.rate=<sample rate> ) "				\
+			"( audio.channels=<number of channels> ) "		\
+			"( audio.position=<channel map> ) "			\
+			"( tunnel.mode=capture|playback|sink|source )"		\
+			"( pipe.filename=<filename> )"				\
+			"( stream.props=<properties> ) "
 
 
 static const struct spa_dict_item module_props[] = {
@@ -157,6 +140,7 @@ static const struct spa_dict_item module_props[] = {
 
 struct impl {
 	struct pw_context *context;
+	struct pw_loop *data_loop;
 
 #define MODE_PLAYBACK	0
 #define MODE_CAPTURE	1
@@ -176,6 +160,8 @@ struct impl {
 	char *filename;
 	unsigned int unlink_fifo;
 	int fd;
+	struct spa_source *socket;
+	struct spa_source *timer;
 
 	struct pw_properties *stream_props;
 	enum pw_direction direction;
@@ -185,10 +171,78 @@ struct impl {
 	uint32_t frame_size;
 
 	unsigned int do_disconnect:1;
-	uint32_t leftover_count;
-	uint8_t *leftover;
+	unsigned int driving:1;
+	unsigned int have_sync:1;
+
+	struct spa_ringbuffer ring;
+	void *buffer;
+	uint32_t target_buffer;
+
+	struct spa_io_rate_match *rate_match;
+	struct spa_io_position *position;
+
+	struct spa_dll dll;
+	float max_error;
+	float corr;
+
+	uint64_t next_time;
 };
 
+static uint64_t get_time_ns(struct impl *impl)
+{
+	struct timespec now;
+	if (spa_system_clock_gettime(impl->data_loop->system, CLOCK_MONOTONIC, &now) < 0)
+		return 0;
+	return SPA_TIMESPEC_TO_NSEC(&now);
+}
+
+static int set_timeout(struct impl *impl, uint64_t time)
+{
+	struct timespec timeout, interval;
+	timeout.tv_sec = time / SPA_NSEC_PER_SEC;
+	timeout.tv_nsec = time % SPA_NSEC_PER_SEC;
+	interval.tv_sec = 0;
+	interval.tv_nsec = 0;
+	pw_loop_update_timer(impl->data_loop,
+                                impl->timer, &timeout, &interval, true);
+	return 0;
+}
+
+static void on_timeout(void *d, uint64_t expirations)
+{
+	struct impl *impl = d;
+	uint64_t duration, current_time;
+	uint32_t rate, index;
+	int32_t avail;
+	struct spa_io_position *pos = impl->position;
+
+	if (SPA_LIKELY(pos)) {
+		duration = pos->clock.target_duration;
+		rate = pos->clock.target_rate.denom;
+	} else {
+		duration = 1024;
+		rate = 48000;
+	}
+	pw_log_debug("timeout %"PRIu64, duration);
+
+	current_time = impl->next_time;
+	impl->next_time += duration / impl->corr * 1e9 / rate;
+	avail = spa_ringbuffer_get_read_index(&impl->ring, &index);
+
+	if (SPA_LIKELY(pos)) {
+                pos->clock.nsec = current_time;
+                pos->clock.rate = pos->clock.target_rate;
+                pos->clock.position += pos->clock.duration;
+                pos->clock.duration = pos->clock.target_duration;
+                pos->clock.delay = SPA_SCALE32_UP(avail, rate, impl->info.rate);
+                pos->clock.rate_diff = impl->corr;
+                pos->clock.next_nsec = impl->next_time;
+        }
+	set_timeout(impl, impl->next_time);
+
+	pw_stream_trigger_process(impl->stream);
+}
+
 static void stream_destroy(void *d)
 {
 	struct impl *impl = d;
@@ -206,8 +260,20 @@ static void stream_state_changed(void *d, enum pw_stream_state old,
 		pw_impl_module_schedule_destroy(impl->module);
 		break;
 	case PW_STREAM_STATE_PAUSED:
+		if (impl->direction == PW_DIRECTION_OUTPUT) {
+			pw_loop_update_io(impl->data_loop, impl->socket, 0);
+			set_timeout(impl, 0);
+		}
 		break;
 	case PW_STREAM_STATE_STREAMING:
+		if (impl->direction == PW_DIRECTION_OUTPUT) {
+			pw_loop_update_io(impl->data_loop, impl->socket, SPA_IO_IN);
+			impl->driving = pw_stream_is_driving(impl->stream);
+			if (impl->driving) {
+				impl->next_time = get_time_ns(impl);
+				set_timeout(impl, impl->next_time);
+			}
+		}
 		break;
 	default:
 		break;
@@ -241,9 +307,12 @@ static void playback_stream_process(void *data)
 					continue;
 				} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
 					/* Don't continue writing */
+					pw_log_debug("pipe (%s) overrun: %m",
+							impl->filename);
 					break;
 				} else {
-					pw_log_warn("Failed to write to pipe sink");
+					pw_log_warn("Failed to write to pipe (%s): %m",
+							impl->filename);
 				}
 			}
 			offs += written;
@@ -253,53 +322,95 @@ static void playback_stream_process(void *data)
 	pw_stream_queue_buffer(impl->stream, buf);
 }
 
+static void update_rate(struct impl *impl, uint32_t filled)
+{
+	float error;
+
+	if (impl->rate_match == NULL)
+		return;
+
+	error = (float)impl->target_buffer - (float)(filled);
+	error = SPA_CLAMP(error, -impl->max_error, impl->max_error);
+
+	impl->corr = spa_dll_update(&impl->dll, error);
+	pw_log_debug("error:%f corr:%f current:%u target:%u",
+			error, impl->corr, filled, impl->target_buffer);
+
+	if (!impl->driving) {
+		SPA_FLAG_SET(impl->rate_match->flags, SPA_IO_RATE_MATCH_FLAG_ACTIVE);
+		impl->rate_match->rate = 1.0f / impl->corr;
+	}
+}
+
 static void capture_stream_process(void *data)
 {
 	struct impl *impl = data;
 	struct pw_buffer *buf;
-	struct spa_data *d;
-	uint32_t req;
-	ssize_t nread;
+	struct spa_data *bd;
+	uint32_t req, index, size;
+	int32_t avail;
 
 	if ((buf = pw_stream_dequeue_buffer(impl->stream)) == NULL) {
-		pw_log_debug("out of buffers: %m");
+		pw_log_warn("out of buffers: %m");
 		return;
 	}
 
-	d = &buf->buffer->datas[0];
+	bd = &buf->buffer->datas[0];
 
 	if ((req = buf->requested * impl->frame_size) == 0)
 		req = 4096 * impl->frame_size;
 
-	req = SPA_MIN(req, d->maxsize);
+	size = SPA_MIN(req, bd->maxsize);
+	size = SPA_ROUND_DOWN(size, impl->frame_size);
 
-	d->chunk->offset = 0;
-	d->chunk->stride = impl->frame_size;
-	d->chunk->size = SPA_MIN(req, impl->leftover_count);
-	memcpy(d->data, impl->leftover, d->chunk->size);
-	req -= d->chunk->size;
+	avail = spa_ringbuffer_get_read_index(&impl->ring, &index);
 
-	nread = read(impl->fd, SPA_PTROFF(d->data, d->chunk->size, void), req);
-	if (nread < 0) {
-		const bool important = !(errno == EINTR
-					 || errno == EAGAIN
-					 || errno == EWOULDBLOCK);
+	pw_log_debug("avail %d %u %u", avail, index, size);
 
-		if (important)
-			pw_log_warn("failed to read from pipe (%s): %s",
-				    impl->filename, strerror(errno));
+	if (avail < (int32_t)size) {
+		memset(bd->data, 0, size);
+		if (avail > 0)
+			pw_log_warn("underrun %d < %u", avail, size);
+		impl->have_sync = false;
 	}
-	else {
-		d->chunk->size += nread;
+	if (avail > (int32_t)RINGBUFFER_SIZE) {
+		index += avail - impl->target_buffer;
+		avail = impl->target_buffer;
+		pw_log_warn("overrun %d > %u", avail, RINGBUFFER_SIZE);
 	}
+	if (avail > 0) {
+		avail = SPA_ROUND_DOWN(avail, impl->frame_size);
+		update_rate(impl, avail);
 
-	impl->leftover_count = d->chunk->size % impl->frame_size;
-	d->chunk->size -= impl->leftover_count;
-	memcpy(impl->leftover, SPA_PTROFF(d->data, d->chunk->size, void), impl->leftover_count);
+		avail = SPA_MIN(size, (uint32_t)avail);
+		spa_ringbuffer_read_data(&impl->ring,
+				impl->buffer, RINGBUFFER_SIZE,
+				index & RINGBUFFER_MASK,
+				bd->data, avail);
+
+		index += avail;
+		spa_ringbuffer_read_update(&impl->ring, index);
+	}
+	bd->chunk->offset = 0;
+	bd->chunk->size = size;
+	bd->chunk->stride = impl->frame_size;
 
 	pw_stream_queue_buffer(impl->stream, buf);
 }
 
+static void stream_io_changed(void *data, uint32_t id, void *area, uint32_t size)
+{
+	struct impl *impl = data;
+	switch (id) {
+	case SPA_IO_RateMatch:
+		impl->rate_match = area;
+		break;
+	case SPA_IO_Position:
+		impl->position = area;
+		break;
+	}
+}
+
 static const struct pw_stream_events playback_stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = stream_destroy,
@@ -310,11 +421,12 @@ static const struct pw_stream_events playback_stream_events = {
 static const struct pw_stream_events capture_stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = stream_destroy,
+	.io_changed = stream_io_changed,
 	.state_changed = stream_state_changed,
-	.process = capture_stream_process
+	.process = capture_stream_process,
 };
 
-static int  create_stream(struct impl *impl)
+static int create_stream(struct impl *impl)
 {
 	int res;
 	uint32_t n_params;
@@ -355,6 +467,92 @@ static int  create_stream(struct impl *impl)
 	return 0;
 }
 
+static inline void
+set_iovec(struct spa_ringbuffer *rbuf, void *buffer, uint32_t size,
+		uint32_t offset, struct iovec *iov, uint32_t len)
+{
+	iov[0].iov_len = SPA_MIN(len, size - offset);
+	iov[0].iov_base = SPA_PTROFF(buffer, offset, void);
+	iov[1].iov_len = len - iov[0].iov_len;
+	iov[1].iov_base = buffer;
+}
+
+static int handle_pipe_read(struct impl *impl)
+{
+	ssize_t nread;
+	int32_t filled;
+	uint32_t index;
+	struct iovec iov[2];
+
+	filled = spa_ringbuffer_get_write_index(&impl->ring, &index);
+	if (!impl->have_sync) {
+		memset(impl->buffer, 0, RINGBUFFER_SIZE);
+	}
+
+	if (filled < 0) {
+		pw_log_warn("%p: underrun write:%u filled:%d",
+				impl, index, filled);
+	}
+
+	set_iovec(&impl->ring,
+			impl->buffer, RINGBUFFER_SIZE,
+			index & RINGBUFFER_MASK,
+			iov, RINGBUFFER_SIZE);
+
+	nread = read(impl->fd, iov[0].iov_base, iov[0].iov_len);
+	if (nread > 0) {
+		index += nread;
+		filled += nread;
+		if (nread == (ssize_t)iov[0].iov_len) {
+			nread = read(impl->fd, iov[1].iov_base, iov[1].iov_len);
+			if (nread > 0) {
+				index += nread;
+				filled += nread;
+			}
+		}
+	}
+	if (!impl->have_sync) {
+		impl->ring.readindex = index - impl->target_buffer;
+
+		spa_dll_init(&impl->dll);
+		spa_dll_set_bw(&impl->dll, SPA_DLL_BW_MIN, 256.f, impl->info.rate);
+		impl->corr = 1.0f;
+
+		pw_log_info("resync");
+		impl->have_sync = true;
+	}
+	spa_ringbuffer_write_update(&impl->ring, index);
+
+	if (nread < 0) {
+		const bool important = !(errno == EINTR
+					 || errno == EAGAIN
+					 || errno == EWOULDBLOCK);
+
+		if (important)
+			pw_log_warn("failed to read from pipe (%s): %m",
+				    impl->filename);
+		else
+			pw_log_debug("pipe (%s) underrun: %m", impl->filename);
+	}
+	pw_log_debug("filled %d %u %d", filled, index, impl->target_buffer);
+
+	return 0;
+}
+
+
+static void on_pipe_io(void *data, int fd, uint32_t mask)
+{
+	struct impl *impl = data;
+
+	if (mask & (SPA_IO_ERR | SPA_IO_HUP)) {
+		pw_log_warn("error:%08x", mask);
+		pw_loop_update_io(impl->data_loop, impl->socket, 0);
+		return;
+	}
+	if (mask & SPA_IO_IN)
+		handle_pipe_read(impl);
+}
+
 static int create_fifo(struct impl *impl)
 {
 	struct stat st;
@@ -383,7 +581,6 @@ static int create_fifo(struct impl *impl)
 
 		do_unlink_fifo = true;
 	}
-
 	if ((fd = open(filename, O_RDWR | O_CLOEXEC | O_NONBLOCK, 0)) < 0) {
 		res = -errno;
 		pw_log_error("open('%s'): %s", filename, spa_strerror(res));
@@ -401,6 +598,20 @@ static int create_fifo(struct impl *impl)
 		pw_log_error("'%s' is not a FIFO.", filename);
 		goto error;
 	}
+	impl->socket = pw_loop_add_io(impl->data_loop, fd,
+			0, false, on_pipe_io, impl);
+	if (impl->socket == NULL) {
+		res = -errno;
+		pw_log_error("can't create socket");
+		goto error;
+	}
+	impl->timer = pw_loop_add_timer(impl->data_loop, on_timeout, impl);
+	if (impl->timer == NULL) {
+		res = -errno;
+		pw_log_error("can't create timer");
+		goto error;
+	}
+
 	pw_log_info("%s fifo '%s' with format:%s channels:%d rate:%d",
 			impl->direction == PW_DIRECTION_OUTPUT ? "reading from" : "writing to",
 			filename,
@@ -410,6 +621,7 @@ static int create_fifo(struct impl *impl)
 	impl->filename = strdup(filename);
 	impl->unlink_fifo = do_unlink_fifo;
 	impl->fd = fd;
+
 	return 0;
 
 error:
@@ -460,13 +672,17 @@ static void impl_destroy(struct impl *impl)
 			unlink(impl->filename);
 		free(impl->filename);
 	}
+	if (impl->socket)
+		pw_loop_destroy_source(impl->data_loop, impl->socket);
+	if (impl->timer)
+		pw_loop_destroy_source(impl->data_loop, impl->timer);
 	if (impl->fd >= 0)
 		close(impl->fd);
 
 	pw_properties_free(impl->stream_props);
 	pw_properties_free(impl->props);
 
-	free(impl->leftover);
+	free(impl->buffer);
 	free(impl);
 }
 
@@ -589,6 +805,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	struct pw_properties *props = NULL;
 	struct impl *impl;
 	const char *str, *media_class = NULL;
+	struct pw_data_loop *data_loop;
 	int res;
 
 	PW_LOG_TOPIC_INIT(mod_topic);
@@ -621,6 +838,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 
 	impl->module = module;
 	impl->context = context;
+	data_loop = pw_context_get_data_loop(context);
+	impl->data_loop = pw_data_loop_get_loop(data_loop);
 
 	if ((str = pw_properties_get(props, "tunnel.mode")) == NULL)
 		str = "playback";
@@ -682,12 +901,18 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 
 	copy_props(impl, props, PW_KEY_NODE_RATE);
 
-	impl->leftover = calloc(1, impl->frame_size);
-	if (impl->leftover == NULL) {
+	impl->buffer = calloc(1, RINGBUFFER_SIZE);
+	if (impl->buffer == NULL) {
 		res = -errno;
-		pw_log_error("can't alloc leftover buffer: %m");
+		pw_log_error("can't alloc ringbuffer: %m");
 		goto error;
 	}
+	spa_ringbuffer_init(&impl->ring);
+	impl->target_buffer = 8192 * impl->frame_size;
+	spa_dll_init(&impl->dll);
+	spa_dll_set_bw(&impl->dll, SPA_DLL_BW_MIN, 256.f, impl->info.rate);
+	impl->max_error = 256.0f * impl->frame_size;
+	impl->corr = 1.0f;
 
 	impl->core = pw_context_get_object(impl->context, PW_TYPE_INTERFACE_Core);
 	if (impl->core == NULL) {
